@@ -2,7 +2,7 @@
  *
  *  OBEX Server
  *
- *  Copyright (C) 2007-2009  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2007-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -42,178 +42,8 @@
 #include "obex.h"
 #include "dbus.h"
 #include "logging.h"
-
-static const gchar *opp_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
-<record>									\
-  <attribute id=\"0x0001\">							\
-    <sequence>									\
-      <uuid value=\"0x1105\"/>							\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0004\">							\
-    <sequence>									\
-      <sequence>								\
-        <uuid value=\"0x0100\"/>						\
-      </sequence>								\
-      <sequence>								\
-        <uuid value=\"0x0003\"/>						\
-        <uint8 value=\"%u\" name=\"channel\"/>					\
-      </sequence>								\
-      <sequence>								\
-        <uuid value=\"0x0008\"/>						\
-      </sequence>								\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0009\">							\
-    <sequence>									\
-      <sequence>								\
-        <uuid value=\"0x1105\"/>						\
-        <uint16 value=\"0x0100\" name=\"version\"/>				\
-      </sequence>								\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0100\">							\
-    <text value=\"%s\" name=\"name\"/>						\
-  </attribute>									\
-										\
-  <attribute id=\"0x0303\">							\
-    <sequence>									\
-      <uint8 value=\"0x01\"/>							\
-      <uint8 value=\"0x01\"/>							\
-      <uint8 value=\"0x02\"/>							\
-      <uint8 value=\"0x03\"/>							\
-      <uint8 value=\"0x04\"/>							\
-      <uint8 value=\"0x05\"/>							\
-      <uint8 value=\"0x06\"/>							\
-      <uint8 value=\"0xff\"/>							\
-    </sequence>									\
-  </attribute>									\
-</record>";
-
-static const gchar *ftp_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
-<record>									\
-  <attribute id=\"0x0001\">							\
-    <sequence>									\
-      <uuid value=\"0x1106\"/>							\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0004\">							\
-    <sequence>									\
-      <sequence>								\
-        <uuid value=\"0x0100\"/>						\
-      </sequence>								\
-      <sequence>								\
-        <uuid value=\"0x0003\"/>						\
-        <uint8 value=\"%u\" name=\"channel\"/>					\
-      </sequence>								\
-      <sequence>								\
-        <uuid value=\"0x0008\"/>						\
-      </sequence>								\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0009\">							\
-    <sequence>									\
-      <sequence>								\
-        <uuid value=\"0x1106\"/>						\
-        <uint16 value=\"0x0100\" name=\"version\"/>				\
-      </sequence>								\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0100\">							\
-    <text value=\"%s\" name=\"name\"/>						\
-  </attribute>									\
-</record>";
-
-static const gchar *pbap_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
-<record>									\
-  <attribute id=\"0x0001\">							\
-    <sequence>									\
-      <uuid value=\"0x112f\"/>							\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0004\">							\
-    <sequence>									\
-      <sequence>								\
-        <uuid value=\"0x0100\"/>						\
-      </sequence>								\
-      <sequence>								\
-        <uuid value=\"0x0003\"/>						\
-        <uint8 value=\"%u\" name=\"channel\"/>					\
-      </sequence>								\
-      <sequence>								\
-        <uuid value=\"0x0008\"/>						\
-      </sequence>								\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0009\">							\
-    <sequence>									\
-      <sequence>								\
-        <uuid value=\"0x1130\"/>						\
-        <uint16 value=\"0x0100\" name=\"version\"/>				\
-      </sequence>								\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0100\">							\
-    <text value=\"%s\" name=\"name\"/>						\
-  </attribute>									\
-										\
-  <attribute id=\"0x0314\">							\
-    <uint8 value=\"0x01\"/>							\
-  </attribute>									\
-</record>";
-
-static const gchar *pcsuite_record =
-"<?xml version=\"1.0\" encoding=\"UTF-8\" ?>					\
-<record>									\
-  <attribute id=\"0x0001\">							\
-    <sequence>									\
-      <uuid value=\"00005005-0000-1000-8000-0002ee000001\"/>			\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0004\">							\
-    <sequence>									\
-      <sequence>								\
-        <uuid value=\"0x0100\"/>						\
-      </sequence>								\
-      <sequence>								\
-        <uuid value=\"0x0003\"/>						\
-        <uint8 value=\"%u\" name=\"channel\"/>					\
-      </sequence>								\
-      <sequence>								\
-        <uuid value=\"0x0008\"/>						\
-      </sequence>								\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0005\">							\
-    <sequence>									\
-      <uuid value=\"0x1002\"/>							\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0009\">							\
-    <sequence>									\
-      <sequence>								\
-        <uuid value=\"00005005-0000-1000-8000-0002ee000001\"/> 			\
-        <uint16 value=\"0x0100\" name=\"version\"/>				\
-      </sequence>								\
-    </sequence>									\
-  </attribute>									\
-										\
-  <attribute id=\"0x0100\">							\
-    <text value=\"%s\" name=\"name\"/>						\
-  </attribute>									\
-</record>";
+#include "btio.h"
+#include "service.h"
 
 #define TRANSFER_INTERFACE OPENOBEX_SERVICE ".Transfer"
 #define SESSION_INTERFACE OPENOBEX_SERVICE ".Session"
@@ -226,14 +56,18 @@ struct agent {
 	gboolean	auth_pending;
 	gchar		*new_name;
 	gchar		*new_folder;
+	guint		watch_id;
 };
 
 static struct agent *agent = NULL;
 
 struct pending_request {
+	DBusPendingCall *call;
 	struct server *server;
-	gchar address[18];
-	gint nsk;
+	gchar *adapter_path;
+	char address[18];
+	guint watch;
+	GIOChannel *io;
 };
 
 struct adapter_any {
@@ -248,6 +82,9 @@ static guint listener_id = 0;
 
 static void agent_free(struct agent *agent)
 {
+	if (!agent)
+		return;
+
 	g_free(agent->new_folder);
 	g_free(agent->new_name);
 	g_free(agent->bus_name);
@@ -383,8 +220,8 @@ static DBusMessage *register_agent(DBusConnection *conn,
 	agent->bus_name = g_strdup(sender);
 	agent->path = g_strdup(path);
 
-	g_dbus_add_disconnect_watch(conn, sender,
-			agent_disconnected, NULL, NULL);
+	agent->watch_id = g_dbus_add_disconnect_watch(conn, sender,
+					agent_disconnected, NULL, NULL);
 
 	debug("Agent registered");
 
@@ -411,12 +248,27 @@ static DBusMessage *unregister_agent(DBusConnection *conn,
 	if (strcmp(agent->bus_name, sender) != 0)
 		return not_authorized(msg);
 
+	g_dbus_remove_watch(conn, agent->watch_id);
+
 	agent_free(agent);
 	agent = NULL;
 
+	debug("Agent unregistered");
+
 	return dbus_message_new_method_return(msg);
 }
 
+static char *target2str(const uint8_t *t)
+{
+	if (!t)
+		return NULL;
+
+	return g_strdup_printf("%02X%02X%02X%02X-%02X%02X-%02X%02X-"
+				"%02X%02X-%02X%02X%02X%02X%02X%02X",
+				t[0], t[1], t[2], t[3], t[4], t[5], t[6],t[7],
+				t[8], t[9], t[10], t[11], t[12], t[13], t[14], t[15]);
+}
+
 static DBusMessage *get_properties(DBusConnection *conn,
 				DBusMessage *msg, void *data)
 {
@@ -424,9 +276,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	DBusMessage *reply;
 	DBusMessageIter iter;
 	DBusMessageIter dict;
-	gchar uuid[37];
-	const gchar *ptr = uuid;
-	const uint8_t *t = os->target;
+	gchar *uuid;
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
@@ -439,12 +289,11 @@ static DBusMessage *get_properties(DBusConnection *conn,
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
 	/* Target */
-	sprintf(uuid, "%02X%02X%02X%02X-%02X%02X-%02X%02X-"
-				"%02X%02X-%02X%02X%02X%02X%02X%02X",
-				t[0], t[1], t[2], t[3], t[4], t[5], t[6],t[7],
-				t[8], t[9], t[10], t[11], t[12], t[13], t[14], t[15]);
+	uuid = target2str(os->service->target);
 	dbus_message_iter_append_dict_entry(&dict, "Target",
-					DBUS_TYPE_STRING, &ptr);
+					DBUS_TYPE_STRING, &uuid);
+	g_free(uuid);
+
 	/* Root folder */
 	dbus_message_iter_append_dict_entry(&dict, "Root",
 					DBUS_TYPE_STRING, &os->server->folder);
@@ -503,32 +352,6 @@ static GDBusMethodTable session_methods[] = {
 	{ }
 };
 
-static gchar *create_xml_record(const char *name,
-			uint16_t service, uint8_t channel)
-{
-	gchar *xml;
-
-	switch (service) {
-	case OBEX_OPP:
-		xml = g_markup_printf_escaped(opp_record, channel, name);
-		break;
-	case OBEX_FTP:
-		xml = g_markup_printf_escaped(ftp_record, channel, name);
-		break;
-	case OBEX_PBAP:
-		xml = g_markup_printf_escaped(pbap_record, channel, name);
-		break;
-	case OBEX_PCSUITE:
-		xml = g_markup_printf_escaped(pcsuite_record, channel, name);
-		break;
-	default:
-		xml = NULL;
-		break;
-	}
-
-	return xml;
-}
-
 static void add_record_reply(DBusPendingCall *call, gpointer user_data)
 {
 	struct server *server = user_data;
@@ -543,13 +366,17 @@ static void add_record_reply(DBusPendingCall *call, gpointer user_data)
 		dbus_error_free(&derr);
 		handle = 0;
 	} else {
+		struct obex_service_driver *driver;
+
 		dbus_message_get_args(reply, NULL,
 				DBUS_TYPE_UINT32, &handle,
 				DBUS_TYPE_INVALID);
 		server->handle = handle;
 
+		driver = (struct obex_service_driver *) server->drivers->data;
+
 		debug("Registered: %s, handle: 0x%x, folder: %s",
-				server->name, handle, server->folder);
+				driver->name, handle, server->folder);
 	}
 
 	dbus_message_unref(reply);
@@ -584,6 +411,7 @@ failed:
 
 void register_record(struct server *server, gpointer user_data)
 {
+	struct obex_service_driver *driver;
 	gchar *xml;
 	gint ret;
 
@@ -596,7 +424,9 @@ void register_record(struct server *server, gpointer user_data)
 		return;
 	}
 
-	xml = create_xml_record(server->name, server->services, server->channel);
+	driver = (struct obex_service_driver *) server->drivers->data;
+	xml = g_markup_printf_escaped(driver->record, driver->channel,
+					driver->name);
 	ret = add_record(any->path, xml, server);
 	g_free(xml);
 }
@@ -604,7 +434,6 @@ void register_record(struct server *server, gpointer user_data)
 static void find_adapter_any_reply(DBusPendingCall *call, gpointer user_data)
 {
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
-	struct server *server;
 	const char *path;
 	gchar *xml;
 	GSList *l;
@@ -625,9 +454,12 @@ static void find_adapter_any_reply(DBusPendingCall *call, gpointer user_data)
 	any->path = g_strdup(path);
 
 	for (l = any->servers; l; l = l->next) {
-		server = l->data;
-		xml = create_xml_record(server->name,
-				server->services, server->channel);
+		struct server *server = l->data;
+		struct obex_service_driver *driver;
+
+		driver = (struct obex_service_driver *) server->drivers->data;
+		xml = g_markup_printf_escaped(driver->record, driver->channel,
+						driver->name);
 		add_record(path, xml, server);
 		g_free(xml);
 	}
@@ -638,7 +470,8 @@ done:
 
 	dbus_message_unref(reply);
 }
-static void find_adapter(const char *pattern,
+
+static DBusPendingCall *find_adapter(const char *pattern,
 				DBusPendingCallNotifyFunction function,
 				gpointer user_data)
 {
@@ -649,27 +482,32 @@ static void find_adapter(const char *pattern,
 
 	msg = dbus_message_new_method_call("org.bluez", "/",
 					"org.bluez.Manager", "FindAdapter");
+	if (!msg)
+		return NULL;
 
 	dbus_message_append_args(msg, DBUS_TYPE_STRING, &pattern,
 			DBUS_TYPE_INVALID);
 
-	dbus_connection_send_with_reply(system_conn, msg, &call, -1);
+	if (!dbus_connection_send_with_reply(system_conn, msg, &call, -1)) {
+		dbus_message_unref(msg);
+		return NULL;
+	}
+
 	dbus_pending_call_set_notify(call, function, user_data, NULL);
-	dbus_pending_call_unref(call);
 
 	dbus_message_unref(msg);
-}
 
-static gboolean find_adapter_any(gpointer user_data)
-{
-	find_adapter("any", find_adapter_any_reply, user_data);
-
-	return FALSE;
+	return call;
 }
 
 static void name_acquired(DBusConnection *conn, void *user_data)
 {
-	find_adapter_any(NULL);
+	DBusPendingCall *call;
+
+	call = find_adapter("any", find_adapter_any_reply, NULL);
+	if (call)
+		dbus_pending_call_unref(call);
+
 	bluetooth_start();
 }
 
@@ -690,7 +528,8 @@ gboolean manager_init(void)
 
 	dbus_error_init(&err);
 
-	connection = g_dbus_setup_bus(DBUS_BUS_SESSION, OPENOBEX_SERVICE, &err);
+	connection = g_dbus_setup_private(DBUS_BUS_SESSION, OPENOBEX_SERVICE,
+									&err);
 	if (connection == NULL) {
 		if (dbus_error_is_set(&err) == TRUE) {
 			fprintf(stderr, "%s\n", err.message);
@@ -701,14 +540,15 @@ gboolean manager_init(void)
 	}
 
 	system_conn = g_dbus_setup_bus(DBUS_BUS_SYSTEM, NULL, NULL);
-	if (system_conn == NULL)
+	if (system_conn == NULL) {
+		dbus_connection_unref(connection);
+		connection = NULL;
 		return FALSE;
+	}
 
 	listener_id = g_dbus_add_service_watch(system_conn, "org.bluez",
 				name_acquired, name_released, NULL, NULL);
 
-	g_idle_add(find_adapter_any, NULL);
-
 	return g_dbus_register_interface(connection, OPENOBEX_MANAGER_PATH,
 					OPENOBEX_MANAGER_INTERFACE,
 					manager_methods, manager_signals, NULL,
@@ -828,11 +668,24 @@ void unregister_transfer(guint32 id)
 	g_free(path);
 }
 
+static void agent_cancel()
+{
+	DBusMessage *msg;
+
+	msg = dbus_message_new_method_call(agent->bus_name, agent->path,
+					"org.openobex.Agent", "Cancel");
+
+	g_dbus_send_message(connection, msg);
+}
+
 static void agent_reply(DBusPendingCall *call, gpointer user_data)
 {
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	const gchar *name;
 	DBusError derr;
+	gboolean *got_reply = user_data;
+
+	*got_reply = TRUE;
 
 	/* Received a reply after the agent exited */
 	if (!agent)
@@ -844,6 +697,10 @@ static void agent_reply(DBusPendingCall *call, gpointer user_data)
 	if (dbus_set_error_from_message(&derr, reply)) {
 		error("Agent replied with an error: %s, %s",
 				derr.name, derr.message);
+
+		if (dbus_error_has_name(&derr, DBUS_ERROR_NO_REPLY))
+			agent_cancel();
+
 		dbus_error_free(&derr);
 		dbus_message_unref(reply);
 		return;
@@ -870,17 +727,8 @@ static void agent_reply(DBusPendingCall *call, gpointer user_data)
 static gboolean auth_error(GIOChannel *io, GIOCondition cond,
 			gpointer user_data)
 {
-	DBusMessage *msg;
-
 	agent->auth_pending = FALSE;
 
-	msg = dbus_message_new_method_call(agent->bus_name, agent->path,
-					"org.openobex.Agent", "Cancel");
-
-	dbus_connection_send_with_reply(connection, msg, NULL, -1);
-
-	dbus_message_unref(msg);
-
 	return FALSE;
 }
 
@@ -897,6 +745,7 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 	const gchar *bda = address;
 	gchar *path;
 	guint watch;
+	gboolean got_reply;
 
 	if (!agent)
 		return -1;
@@ -940,6 +789,7 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 	dbus_message_unref(msg);
 
 	agent->auth_pending = TRUE;
+	got_reply = FALSE;
 
 	/* Catches errors before authorization response comes */
 	io = g_io_channel_unix_new(fd);
@@ -948,7 +798,7 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 			auth_error, NULL, NULL);
 	g_io_channel_unref(io);
 
-	dbus_pending_call_set_notify(call, agent_reply, NULL, NULL);
+	dbus_pending_call_set_notify(call, agent_reply, &got_reply, NULL);
 
 	/* Workaround: process events while agent doesn't reply */
 	while (agent && agent->auth_pending)
@@ -956,7 +806,11 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 
 	g_source_remove(watch);
 
-	dbus_pending_call_cancel(call);
+	if (!got_reply) {
+		dbus_pending_call_cancel(call);
+		agent_cancel();
+	}
+
 	dbus_pending_call_unref(call);
 
 	if (!agent || !agent->new_name)
@@ -970,47 +824,87 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 	return 0;
 }
 
+static void service_cancel(struct pending_request *pending)
+{
+	DBusMessage *msg;
+
+	msg = dbus_message_new_method_call("org.bluez",
+					pending->adapter_path,
+					"org.bluez.Service",
+					"CancelAuthorization");
+
+	g_dbus_send_message(system_conn, msg);
+}
+
+void obex_connect_cb(GIOChannel *io, GError *err, gpointer user_data)
+{
+	struct server *server = user_data;
+
+	if (err) {
+		error("%s", err->message);
+		g_io_channel_shutdown(io, TRUE, NULL);
+		return;
+	}
+
+	if (obex_session_start(io, server) < 0)
+		g_io_channel_shutdown(io, TRUE, NULL);
+}
+
+static void pending_request_free(struct pending_request *pending)
+{
+	if (pending->call)
+		dbus_pending_call_unref(pending->call);
+	g_io_channel_unref(pending->io);
+	g_free(pending->adapter_path);
+	g_free(pending);
+}
+
 static void service_reply(DBusPendingCall *call, gpointer user_data)
 {
 	struct pending_request *pending = user_data;
+	GIOChannel *io = pending->io;
+	struct server *server = pending->server;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	DBusError derr;
+	GError *err = NULL;
 
 	dbus_error_init(&derr);
 	if (dbus_set_error_from_message(&derr, reply)) {
 		error("RequestAuthorization error: %s, %s",
 				derr.name, derr.message);
+
+		if (dbus_error_has_name(&derr, DBUS_ERROR_NO_REPLY))
+			service_cancel(pending);
+
 		dbus_error_free(&derr);
-		close(pending->nsk);
+		g_io_channel_shutdown(io, TRUE, NULL);
 		goto done;
 	}
 
 	debug("RequestAuthorization succeeded");
 
-	if (obex_session_start(pending->nsk, pending->server) < 0)
-		close(pending->nsk);
+	if (!bt_io_accept(io, obex_connect_cb, server, NULL, &err)) {
+		error("%s", err->message);
+		g_error_free(err);
+		g_io_channel_shutdown(io, TRUE, NULL);
+	}
 
 done:
-	g_free(pending);
+	g_source_remove(pending->watch);
+	pending_request_free(pending);
 	dbus_message_unref(reply);
 }
 
 static gboolean service_error(GIOChannel *io, GIOCondition cond,
 			gpointer user_data)
 {
-	DBusMessage *msg;
+	struct pending_request *pending = user_data;
 
-	if (any->path == NULL)
-		return FALSE;
+	service_cancel(pending);
 
-	msg = dbus_message_new_method_call("org.bluez",
-					any->path,
-					"org.bluez.Service",
-					"CancelAuthorization");
+	dbus_pending_call_cancel(pending->call);
 
-	dbus_connection_send_with_reply(system_conn, msg, NULL, -1);
-
-	dbus_message_unref(msg);
+	pending_request_free(pending);
 
 	return FALSE;
 }
@@ -1021,9 +915,7 @@ static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	DBusMessage *msg;
 	DBusPendingCall *pcall;
-	GIOChannel *io;
-	const char *paddr = pending->address;
-	const char *path;
+	const char *path, *paddr = pending->address;
 	DBusError derr;
 
 	dbus_error_init(&derr);
@@ -1031,7 +923,7 @@ static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
 		error("Replied with an error: %s, %s",
 				derr.name, derr.message);
 		dbus_error_free(&derr);
-		goto done;
+		goto failed;
 	}
 
 	dbus_message_get_args(reply, NULL,
@@ -1039,6 +931,8 @@ static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
 			DBUS_TYPE_INVALID);
 
 	debug("FindAdapter -> %s", path);
+	pending->adapter_path = g_strdup(path);
+	dbus_message_unref(reply);
 
 	msg = dbus_message_new_method_call("org.bluez", path,
 			"org.bluez.Service", "RequestAuthorization");
@@ -1050,62 +944,64 @@ static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
 	if (!dbus_connection_send_with_reply(system_conn,
 					msg, &pcall, TIMEOUT)) {
 		dbus_message_unref(msg);
-		goto done;
+		goto failed;
 	}
 
 	dbus_message_unref(msg);
 
 	debug("RequestAuthorization(%s, %x)", paddr, pending->server->handle);
 
-	if (!dbus_pending_call_set_notify(pcall, service_reply, pending, NULL)) {
-		close(pending->nsk);
-		g_free(pending);
-		goto done;
+	if (!dbus_pending_call_set_notify(pcall, service_reply, pending,
+								NULL)) {
+		dbus_pending_call_unref(pcall);
+		goto failed;
 	}
 
-	dbus_pending_call_unref(pcall);
+	dbus_pending_call_unref(pending->call);
+	pending->call = pcall;
 
 	/* Catches errors before authorization response comes */
-	io = g_io_channel_unix_new(pending->nsk);
-	g_io_add_watch_full(io, G_PRIORITY_DEFAULT,
-			G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-			service_error, NULL, NULL);
-	g_io_channel_unref(io);
+	pending->watch = g_io_add_watch(pending->io,
+					G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+					service_error, pending);
 
-done:
-	dbus_message_unref(reply);
+	return;
+
+failed:
+	g_io_channel_shutdown(pending->io, TRUE, NULL);
+	pending_request_free(pending);
 }
 
-gint request_service_authorization(struct server *server, gint nsk)
+gint request_service_authorization(struct server *server, GIOChannel *io,
+					const char *address)
 {
-	struct sockaddr_rc src, dst;
-	socklen_t addrlen;
-	gchar address[18];
-	const gchar *paddr = address;
 	struct pending_request *pending;
+	char source[18];
+	GError *err = NULL;
 
 	if (system_conn == NULL || any->path == NULL)
 		return -1;
 
-	memset(&src, 0, sizeof(src));
-	memset(&dst, 0, sizeof(dst));
-	addrlen = sizeof(src);
-
-	if (getsockname(nsk, (struct sockaddr *) &src, &addrlen) < 0)
-		return -1;
-
-	ba2str(&src.rc_bdaddr, address);
-
-	if (getpeername(nsk, (struct sockaddr *) &dst, &addrlen) < 0)
-		return -1;
+	bt_io_get(io, BT_IO_RFCOMM, &err,
+			BT_IO_OPT_SOURCE, source,
+			BT_IO_OPT_INVALID);
+	if (err) {
+		error("%s", err->message);
+		g_error_free(err);
+		return -EINVAL;
+	}
 
 	pending = g_new0(struct pending_request, 1);
+	pending->call = find_adapter(source, find_adapter_reply, pending);
+	if (!pending->call) {
+		g_free(pending);
+		return -ENOMEM;
+	}
+
 	pending->server = server;
-	pending->nsk = nsk;
+	pending->io = g_io_channel_ref(io);
+	memcpy(pending->address, address, sizeof(pending->address));
 
-	ba2str(&dst.rc_bdaddr, pending->address);
-
-	find_adapter(paddr, find_adapter_reply, pending);
 
 	return 0;
 }
@@ -1135,3 +1031,11 @@ void unregister_session(guint32 id)
 
 	g_free(path);
 }
+
+DBusConnection *obex_dbus_get_connection(void)
+{
+	if (connection == NULL)
+		return NULL;
+
+	return dbus_connection_ref(connection);
+}
